#include <stdio.h>
#include <stdlib.h>

#include "tbx_timer.h"
#include "matrix.h"

#define  MACHINE_MAIN
#include "comm.h"

#define  MATRIX_SIZE   352

int grid_to_global(int P, int Q, int p, int q)
{
   return(p * P + q);
}

void distribute(matrix *Mat, int M, int N)
{
   int P;
   int Q;
   int p;
   int q;

   int my_address;

   Mat -> num_nodes = machine_info.global.number_of_tasks;
   Mat -> my_address = my_address = machine_info.grid.my_address;

   Mat -> P = P = machine_info.grid.dimension_rank[0];
   Mat -> Q = Q = machine_info.grid.dimension_rank[1];

   Mat -> p = p = machine_info.grid.Position[0];
   Mat -> q = q = machine_info.grid.Position[1];

   Mat -> M = M;
   Mat -> N = N;

   Mat -> m = (M / P) + (p < (M % P));
   Mat -> n = (N / Q) + (q < (N % Q));
#if 0
   printf("%d: (p, q) = (%d, %d), (m, n) = (%d, %d)\n", my_address, p, q,
          Mat -> m, Mat -> n);
#endif
}

void make_matrix(matrix *Mat, int M, int N)
{
   double **Matrix;
   double  *Store;
   int      i;
   int      Index;
   int      m;
   int      n;

   distribute(Mat, M, N);

   m = Mat -> m;
   n = Mat -> n;

   if ((Matrix = (double **) malloc(m * sizeof(double *))) != NULL)
   {
      if ((Store = (double *) malloc(m * n * sizeof(double))) != NULL)
      {
         for (i = 0, Index = 0; i < m; i++, Index += n)
            Matrix[i] = &(Store[Index]);
      }
      else
      {
         printf("Out of memory!\n");
         exit(1);
      }
   }
   else
   {
      printf("Out of memory!\n");
      exit(1);
   }

   Mat -> Matrix = Matrix;
}

void fill_matrix(matrix *Mat)
{
   int      m = Mat -> m;
   int      n = Mat -> n;
   double **Matrix = Mat -> Matrix;
   int      i;
   int      j;

#if 0
   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = (double) random() / MAX_INT;
#endif

#if 1
   if (Mat -> p == Mat -> q)
   {
      for (i = 0; i < m; i++)
         Matrix[i][i] = 1.0;
   }
   else
   {
      for (i = 0; i < m; i++)
         Matrix[i][i] = 0.0;
   }
#endif

#if 0
   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = 1.0;
#endif
}

void clear_matrix(matrix *Mat)
{
   int      m = Mat -> m;
   int      n = Mat -> n;
   double **Matrix = Mat -> Matrix;
   int      i;
   int      j;

   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = 0.0;
}

void mult_local_matrix(int m, int n, double **a, double **b, double **c)
{
   int i;
   int j;
   int k;
   double a_ij;
   double *c_i;
   double *b_j;

   for (i = 0; i < m; i++)
   {
      c_i = c[i];
      for (j = 0; j < n; j++)
      {
         a_ij = a[i][j];
         b_j = b[j];
         for (k = 0; k < m; k++)
            c_i[k] += a_ij * b_j[k];
      }
   }
}

#if 1
#define MPI
#endif
void matrix_mult(matrix *A, matrix *B, matrix *C, matrix *temp_Storage1,
                 matrix *temp_Storage2, matrix *temp_Storage3)
{
   int      q = A -> q;
   int      Q = A -> Q;
   int      p = A -> p;
   int      P = A -> P;
   int      m = A -> m;
   int      n = A -> n;
   int      k;
   int      j;
   int      qhat;
   int      phat;

   int      coordinate[2];

   MPI_Request a_request;
   MPI_Status  a_status;
   int      Boardcast_src;
   int      Boardcast_dest;
   int      Boardcast_tag;

   int      Buffer_size;

   MPI_Request b_request;
   MPI_Status  b_status;
   int      Shift_dest;
   int      Shift_src;
   int      Shift_tag;

   double **a = A -> Matrix;
   double **b = B -> Matrix;
   double **c = C -> Matrix;
   double **temp_a1 = temp_Storage1 -> Matrix;
   double **temp_a2 = temp_Storage2 -> Matrix;
   int      free_temp_a;

   double **working_a;
   double **Buffer_a;

   double **temp_b1 = b;
   double **temp_b2 = temp_Storage3 -> Matrix;
   int      free_temp_b;

   double **working_b;
   double **Buffer_b;

   qhat = phat = A -> p;
   Buffer_size = m * n * sizeof(double);

   Boardcast_tag = A_TAG;

   if (q == qhat)
   {
      coordinate[0] = p;
      /* boardcast the local matrix  */
      for (j = 1; j < Q; j++)
      {
         coordinate[1] = (qhat + j) % Q;
         MPI_Cart_rank(machine_info.grid.comm, coordinate, &Boardcast_dest);
         MPE_Log_send(Boardcast_dest, Boardcast_tag, Buffer_size);
         MPE_Log_event(1,0,"Async. Send");
#ifdef MPI
         MPI_Isend(a[0], Buffer_size, MPI_BYTE, Boardcast_dest, Boardcast_tag,
                   machine_info.grid.comm, &a_request);
#else
         printf("k = 0: (%d, %d) -- sending out local A matrix to (%d, %d).\n",
                p, q, p, (qhat + j) % Q);
#endif
         MPE_Log_event(2,0,"Async. Send");
      }
      free_temp_a = 1;
   }
   else
   {
      /* boardcast the local matrix  */
      coordinate[0] = p;
      coordinate[1] = qhat;
      MPI_Cart_rank(machine_info.grid.comm, coordinate, &Boardcast_src);
      MPE_Log_event(3,0,"Async. Receive");
#ifdef MPI
      MPI_Irecv(temp_a1[0], Buffer_size, MPI_BYTE, Boardcast_src,
                Boardcast_tag, machine_info.grid.comm, &a_request);
#else
      printf("k = 0: (%d, %d) -- receiving local A matrix from (%d, %d).\n",
             p, q, p, qhat);
#endif
      MPE_Log_event(4,0,"Async. Receive");
      free_temp_a = 2;
   }

#if 0
   printf("shift addresss:\n");
#endif
   phat = p - 1;
   if (phat < 0)
      phat = P - 1;
#if 0
   printf("my address = (%d, %d), b send to = (%d, %d)\n", p, q, phat, q);
#endif
   coordinate[0] = phat;
   coordinate[1] = q;
   MPI_Cart_rank(machine_info.grid.comm, coordinate, &Shift_dest);

   phat = (p + 1) % P;
#if 0
   printf("my address = (%d, %d), b receive from = (%d, %d)\n", p, q, phat, q);
#endif
   coordinate[0] = phat;
   coordinate[1] = q;
   MPI_Cart_rank(machine_info.grid.comm, coordinate, &Shift_src);

   free_temp_b = 2;
   working_b = temp_b1;
   Buffer_b = temp_b2;
   Shift_tag = B_TAG;

   k = 0;
   do {
      if (q == qhat)
      {
         working_a = a;
         Buffer_a = temp_a1;
         free_temp_a = 2;
      }
      else
      {
         if (free_temp_a == 2)
         {
            working_a = temp_a1;
            Buffer_a = temp_a2;
         }
         else
         {
            working_a = temp_a2;
            Buffer_a = temp_a1;
         }
      }

      if (free_temp_b == 1)
      {
         working_b = temp_b2;
         Buffer_b = temp_b1;
         free_temp_b = 2;
      }
      else
      {
         working_b = temp_b1;
         Buffer_b = temp_b2;
         free_temp_b = 1;
      }

      if (k > 0)
      {
         /* Waiting for B for the current stage */
         MPE_Log_event(11,0,"Async. Wait");
#ifdef MPI
         MPI_Wait(&b_request, &b_status);
#else
         printf("my address = (%d, %d) waiting for b\n", p, q, phat, q);
#endif
         MPE_Log_event(12,0,"Async. Wait");
         MPE_Log_receive(Shift_src, Shift_tag, Buffer_size);

      }

      if (q != qhat)
      {
         /* Waiting for A for the current stage */
         MPE_Log_event(5,0,"Async. Wait");
#ifdef MPI
         MPI_Wait(&a_request, &a_status);
#else
         printf("my address = (%d, %d) waiting for a\n", p, q, phat, q);
#endif
         MPE_Log_event(6,0,"Async. Wait");
         MPE_Log_receive(Boardcast_src, Boardcast_tag, Buffer_size);
      }

      k++;
      /* Send out the data for next stage */
      if (k < P)
      {
         /* Put the local B matrix on the pipeline */
         Shift_tag = B_TAG + k;
         MPE_Log_send(Shift_dest, Shift_tag, Buffer_size);
         MPE_Log_event(7,0,"Async. Send");
#ifdef MPI
         MPI_Isend(working_b[0], Buffer_size, MPI_BYTE, Shift_dest, Shift_tag,
                   machine_info.grid.comm, &b_request);
#else
         printf("my address = (%d, %d) sending b\n", p, q); 
#endif
         MPE_Log_event(8,0,"Async. Send");

         MPE_Log_event(9,0,"Async. Receive");
#ifdef MPI
         MPI_Irecv(Buffer_b[0], Buffer_size, MPI_BYTE, Shift_src, Shift_tag,
                   machine_info.grid.comm, &b_request);
#else
         printf("my address = (%d, %d) receiving b\n", p, q); 
#endif
         MPE_Log_event(10,0,"Async. Receive");

         qhat = (qhat + 1) % Q;
         Boardcast_tag = A_TAG + k;

         if (q == qhat)
         {
            /* boardcast the local A matrix  */
            for (j = 1; j < Q; j++)
            {
               coordinate[0] = p;
               coordinate[1] = (qhat + j) % Q;
               MPI_Cart_rank(machine_info.grid.comm, coordinate,
                             &Boardcast_dest);
               MPE_Log_send(Boardcast_dest, Boardcast_tag, Buffer_size);
               MPE_Log_event(1,0,"Async. Send");
#ifdef MPI
               MPI_Isend(a[0], Buffer_size, MPI_BYTE, Boardcast_dest,
                         Boardcast_tag, machine_info.grid.comm, &a_request);
#else
               printf("k = %d: (%d, %d) -- sending out local A matrix to (%d, %d).\n",
                      k, p, q, p, (qhat + j) % Q);
#endif
               MPE_Log_event(2,0,"Async. Send");
            }
         }
         else
         {
            /* boardcast the local matrix  */
            coordinate[0] = p;
            coordinate[1] = qhat;
            MPI_Cart_rank(machine_info.grid.comm, coordinate, &Boardcast_src);
            MPE_Log_event(3,0,"Async. Receive");
#ifdef MPI
            MPI_Irecv(Buffer_a[0], Buffer_size, MPI_BYTE, Boardcast_src,
                      Boardcast_tag, machine_info.grid.comm, &a_request);
#else
            printf("k = %d: (%d, %d) -- receiving local A matrix from (%d, %d).\n",
                   k, p, q, p, qhat);
#endif
            MPE_Log_event(4,0,"Async. Receive");
            if (free_temp_a == 1)
               free_temp_a = 2;
            else
               free_temp_a = 1;
         }
      }
      else if (k == P)
      {
         /* Put the local B matrix back to where it is started */
         Shift_tag = B_TAG + k;
         MPE_Log_send(Shift_dest, Shift_tag, Buffer_size);
         MPE_Log_event(7,0,"Async. Send");
#ifdef MPI
         MPI_Isend(working_b[0], Buffer_size, MPI_BYTE, Shift_dest, Shift_tag,
                  machine_info.grid.comm, &b_request);
#else
         printf("my address = (%d, %d) sending b\n", p, q);
#endif
         MPE_Log_event(8,0,"Async. Send");

         MPE_Log_event(9,0,"Async. Receive");
#ifdef MPI
         MPI_Irecv(b[0], Buffer_size, MPI_BYTE, Shift_src, Shift_tag,
                   machine_info.grid.comm, &b_request);
#else
         printf("my address = (%d, %d) receiving b\n", p, q);
#endif
         MPE_Log_event(10,0,"Async. Receive");
      }

      /* Compute A * B for k - 1 iteration */
#if 0
      printf("mult_local_matrix\n");
#endif
      mult_local_matrix(m, n, working_a, working_b, c);
#if 0
      printf("(%d, %d) -- shifting local B matrix.\n", B -> p, B -> q);
#endif
   } while (k < P);

   /* Waiting for B for the current stage */
   MPE_Log_event(11,0,"Async. Wait");
#ifdef MPI
   MPI_Wait(&b_request, &b_status);
#else
   printf("my address = (%d, %d) waiting for b\n", p, q, phat, q);
#endif
   MPE_Log_event(12,0,"Async. Wait");
   MPE_Log_receive(Shift_src, Shift_tag, Buffer_size);

#if 0
   if (p == q)
   {
      for (j = 0; j < m; j++)
         printf("%lf\n", c[j][j]);
   }
#endif
}
#undef MPI

int *set_task_list(int start_task_id, int end_task_id, int step,
                   int number_of_tasks)
{
   int *tasks_list;
   int  id;
   int  i;

   if ((tasks_list = (int *) malloc(number_of_tasks * sizeof(int))) == NULL)
   {
      printf("set_task_list(): Out of memory!\n");
      exit(1);
   }

   for (i = number_of_tasks - 1, id = end_task_id; i >= 0; i--, id -= step)
      tasks_list[i] = id;

#ifdef DEBUG_set_task_list
   printf("start id = %d, end id = %d, number_of_tasks = %d\n",
          id + step , end_task_id, number_of_tasks);
#endif

   return(tasks_list);
}

main(int argc, char *argv[])
{
   int      i;
   matrix   A;
   matrix   B;
   matrix   C;
   matrix   D;
   matrix   E;
   matrix   F;
   unsigned Seed;
   char     State[256];
   double   Total_Opt_per_node;
   tbx_timer Timer;

   MPI_Init(&argc, &argv);

   MPI_Comm_size(MPI_COMM_WORLD, &(machine_info.global.number_of_tasks));
   MPI_Comm_rank(MPI_COMM_WORLD, &(machine_info.global.my_address));
   MPI_Comm_group(MPI_COMM_WORLD, &(machine_info.global.group));
   machine_info.global.comm = MPI_COMM_WORLD;

   machine_info.grid.dimension = 2;
   
   if ((machine_info.grid.dimension_rank
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   if ((machine_info.grid.periods
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   if ((machine_info.grid.Position
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   for (i = machine_info.grid.dimension - 1; i >= 0; i--)
   {
      machine_info.grid.dimension_rank[i] = 0;
      machine_info.grid.periods[i] = 1;
      machine_info.grid.Position[i] = 0;
   }

   MPI_Dims_create(machine_info.global.number_of_tasks,
                   machine_info.grid.dimension,
                   machine_info.grid.dimension_rank);
   MPI_Cart_create(MPI_COMM_WORLD, machine_info.grid.dimension,
                   machine_info.grid.dimension_rank,
                   machine_info.grid.periods, 1,
                   &(machine_info.grid.comm));
   MPI_Comm_group(machine_info.grid.comm, &(machine_info.grid.group));
   MPI_Comm_size(machine_info.grid.comm,
                 &(machine_info.grid.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.comm, &(machine_info.grid.my_address));
   MPI_Cart_coords(machine_info.grid.comm, machine_info.grid.my_address,
                   machine_info.grid.dimension, machine_info.grid.Position);

   /* Building rows */
   MPI_Comm_split(machine_info.grid.comm, machine_info.grid.Position[0],
                  0, &(machine_info.grid.row.comm));
   MPI_Comm_group(machine_info.grid.row.comm,
                  &(machine_info.grid.row.group));
   MPI_Comm_group(machine_info.grid.row.comm, &(machine_info.grid.row.group));
   MPI_Comm_size(machine_info.grid.row.comm,
                 &(machine_info.grid.row.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.row.comm,
                 &(machine_info.grid.row.my_address));

   /* Building columns */
   MPI_Comm_split(machine_info.grid.comm, machine_info.grid.Position[1],
                  0, &(machine_info.grid.col.comm));
   MPI_Comm_group(machine_info.grid.col.comm,
                  &(machine_info.grid.col.group));
   MPI_Comm_group(machine_info.grid.col.comm, &(machine_info.grid.col.group));
   MPI_Comm_size(machine_info.grid.col.comm,
                 &(machine_info.grid.col.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.col.comm,
                 &(machine_info.grid.col.my_address));

   MPE_Init_log();
   if (machine_info.global.my_address == 0)
   {
      MPE_Describe_state(1, 2, "A Send", "blue:vlines3");
      MPE_Describe_state(3, 4, "A Receive", "red:gray3");
      MPE_Describe_state(5, 6, "A wait", "green:light_gray");
      MPE_Describe_state(7, 8, "B Send", "AliceBlue");
      MPE_Describe_state(9, 10, "B Receive", "LightBlue");
      MPE_Describe_state(11, 12, "B wait", "IndianRed");
   }
 
   make_matrix(&A, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&A);
   fill_matrix(&A);

   make_matrix(&B, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&B);
   fill_matrix(&B);

   make_matrix(&C, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&C);

   make_matrix(&D, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&D);
   make_matrix(&E, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&E);
   make_matrix(&F, MATRIX_SIZE, MATRIX_SIZE);
   clear_matrix(&F);

   tbx_timer_reset(&Timer);

   MPE_Start_log();
   MPI_Barrier(machine_info.grid.comm);

   tbx_timer_start(&Timer);

   matrix_mult(&A, &B, &C, &D, &E, &F);

   MPI_Barrier(machine_info.grid.comm);
   tbx_timer_stop(&Timer);
   MPE_Stop_log();

   compute_time(&Timer);
   Total_Opt_per_node = ((double) MATRIX_SIZE * MATRIX_SIZE * MATRIX_SIZE)
                        / machine_info.grid.number_of_tasks;

   if (machine_info.global.my_address == 0)
   {
      double Time;

      Time = Timer.total_sec + (Timer.total_usec / 1000000.0);
      printf("%d x %d\n", MATRIX_SIZE, MATRIX_SIZE);
      printf("Time=%lf [secs], flops per sec per node =%lf\n",
             Time, Total_Opt_per_node / Time);
   }

   MPE_Finish_log("mm.log");

   MPI_Finalize();
   exit(0);
}

